from typing import Union

from ._base import CompileTimeValue
from .nbt import CplNBT
from .score import CplScore
from ..tokenizer import Token
from ..utils import STRING_TYPE, get_uuid


class CplStringNBT(CplNBT):
    def __init__(self, token: Union[Token, None], location: str):
        super().__init__(token, location, STRING_TYPE)

    def _get_index(self, ctx, index: CompileTimeValue):
        if isinstance(index, CplString) and index.value == "length":
            eid = f"int_{get_uuid()} __temp__"
            ctx.file.append(f"execute store result score {eid} run data get {self.location}")
            return CplScore(self.token, eid)
        return None

    def _set_add(self, ctx, cpl):
        if isinstance(cpl, CplString):
            file_name = ctx.transpiler.get_temp_file_name(
                f"$data modify {self.location} set value '$(_0){cpl.value}'")
            self.cache(ctx, nbt_loc="string_concat _0", force="nbt")
            ctx.file.append(f"function {ctx.transpiler.pack_namespace}:{file_name} with storage string_concat")
            return self
        if not isinstance(cpl, CplStringNBT):
            file_name = ctx.transpiler.get_temp_file_name(
                f"$data modify {self.location} set value '$(_0)$(_1)'")
            self.cache(ctx, nbt_loc="string_concat _0", force="nbt")
            cpl.cache(ctx, nbt_loc="string_concat _1", force="nbt")
            ctx.file.append(f"function {ctx.transpiler.pack_namespace}:{file_name} with storage string_concat")
            return self
        return None

    def _add(self, ctx, cpl):
        if isinstance(cpl, CplString):
            eid = f"storage temp _{get_uuid()}"
            file_name = ctx.transpiler.get_temp_file_name(
                f"$data modify {eid} set value '$(_0){cpl.value}'")
            self.cache(ctx, nbt_loc="string_concat _0", force="nbt")
            ctx.file.append(f"function {ctx.transpiler.pack_namespace}:{file_name} with storage string_concat")
            return CplStringNBT(self.token, eid)
        if not isinstance(cpl, CplStringNBT):
            eid = f"storage temp _{get_uuid()}"
            file_name = ctx.transpiler.get_temp_file_name(
                f"$data modify {eid} set value '$(_0)$(_1)'")
            self.cache(ctx, nbt_loc="string_concat _0", force="nbt")
            cpl.cache(ctx, nbt_loc="string_concat _1", force="nbt")
            ctx.file.append(f"function {ctx.transpiler.pack_namespace}:{file_name} with storage string_concat")
            return CplStringNBT(self.token, eid)
        return None

    def _and(self, ctx, cpl):
        return cpl._and(ctx, self.get_index(ctx, CplString(self.token, "length"), self.token))

    def _or(self, ctx, cpl):
        return cpl._or(ctx, self.get_index(ctx, CplString(self.token, "length"), self.token))


from .string import CplString



from typing import List

from ..cpl.score import CplScore
from ..tokenizer import GroupToken, Token
from ..transpiler import FunctionDeclaration, TranspilerContext, add_lib
from ..utils import TokenType


def lib_success(ctx: TranspilerContext, _args: List[GroupToken], token: GroupToken):
    c = token.value[8:-1]
    ctx.transpiler.run_cmd(ctx, Token(c, TokenType.POINTER, 0, len(c)), "int_success __temp__", "success")


add_lib(FunctionDeclaration(
    type="python-raw",
    name="success",
    returns=CplScore(None, "int_success __temp__"),
    function=lib_success
))



from .float import CplFloat
from .int import CplInt
from .nbtfloat import CplFloatNBT
from .nbtint import CplIntNBT
from .score import CplScore
from ..error import raise_syntax_error
from ..utils import get_uuid, basic_cmp, FLOAT_PREC, inv_cmp


def num_add(self, ctx, cpl):
    if isinstance(cpl, CplInt) or isinstance(cpl, CplFloat):
        if isinstance(self, CplInt) and isinstance(cpl, CplInt):
            return CplInt(self.token, self.value + cpl.value)
        return CplFloat(self.token, self.value + cpl.value)
    if self.value == 0:
        return cpl
    return self.cache(ctx)._set_add(ctx, cpl)


def num_sub(self, ctx, cpl):
    if isinstance(cpl, CplInt) or isinstance(cpl, CplFloat):
        if isinstance(self, CplInt) and isinstance(cpl, CplInt):
            return CplInt(self.token, self.value - cpl.value)
        return CplFloat(self.token, self.value - cpl.value)
    return self.cache(ctx)._set_sub(ctx, cpl)


def num_mul(self, ctx, cpl):
    if isinstance(cpl, CplInt) or isinstance(cpl, CplFloat):
        if isinstance(self, CplInt) and isinstance(cpl, CplInt):
            return CplInt(self.token, self.value * cpl.value)
        return CplFloat(self.token, self.value * cpl.value)
    if self.value == 0:
        return self
    if self.value == 1:
        return cpl
    return self.cache(ctx)._set_mul(ctx, cpl)


def num_div(self, ctx, cpl):
    if isinstance(cpl, CplInt) or isinstance(cpl, CplFloat):
        if isinstance(self, CplInt) and isinstance(cpl, CplInt):
            return CplInt(self.token, self.value / cpl.value)
        return CplFloat(self.token, self.value / cpl.value)
    if self.value == 0:
        raise_syntax_error("Division by zero", self.token)
    return self.cache(ctx)._set_div(ctx, cpl)


def num_mod(self, ctx, cpl):
    if isinstance(cpl, CplInt) or isinstance(cpl, CplFloat):
        if isinstance(self, CplInt) and isinstance(cpl, CplInt):
            return CplInt(self.token, self.value % cpl.value)
        return CplFloat(self.token, self.value % cpl.value)
    if self.value == 0:
        raise_syntax_error("Modulo by zero", self.token)
    return self.cache(ctx)._set_mod(ctx, cpl)


def num_pow(self, ctx, cpl):
    if isinstance(cpl, CplInt) or isinstance(cpl, CplFloat):
        if isinstance(self, CplInt):
            return CplInt(self.token, self.value ** cpl.value)
        return CplFloat(self.token, self.value ** cpl.value)
    return self.cache(ctx)._set_pow(ctx, cpl)


def num_eq_neq(self, ctx, cpl, is_eq):
    if isinstance(cpl, CplInt) or isinstance(cpl, CplFloat):
        v = float(cpl.value) == float(self.value)
        if not is_eq:
            v = not v
        return CplInt(self.token, v)

    if isinstance(cpl, CplIntNBT) or isinstance(cpl, CplFloatNBT):
        cpl = cpl.cache(ctx, force="score", force_t=self.unique_type.type)

    if isinstance(cpl, CplScore):
        v = self.value
        self_t = self.unique_type.type
        score_t = cpl.unique_type.type
        if score_t == "int" and self_t == "float":
            v = int(v)
        if score_t == "float":
            v = int(v * FLOAT_PREC)
        eid = f"int_{get_uuid()} __temp__"
        ctx.file.append(f"scoreboard players set {eid} 0")
        ctx.file.append(f"execute if score {cpl.location} matches {v}..{v} "
                        f"run scoreboard players set {eid} 1")
        return CplScore(self.token, eid)
    return None


def num_cmp(self, ctx, cpl, op):
    if isinstance(cpl, CplInt) or isinstance(cpl, CplFloat):
        return CplInt(self.token, basic_cmp(float(self.value), op, float(cpl.value)))
    if isinstance(cpl, CplIntNBT) or isinstance(cpl, CplFloatNBT):
        cpl = cpl.cache(ctx, force="score")
    if not isinstance(cpl, CplScore):
        return None
    self_t = self.unique_type.type
    score_t = cpl.unique_type.type
    inv_op = inv_cmp(op)
    # score INV_OP self
    v = self.value
    if score_t == "int" and self_t == "float":
        v = int(v)
    if score_t == "float":
        v = int(v * FLOAT_PREC)
    eid = f"int_{get_uuid()} __temp__"
    ctx.file.append(f"scoreboard players set {eid} 0")
    matcher = ""
    if inv_op == ">":
        matcher = f"{v + 1}.."
    elif inv_op == "<":
        matcher = f"..{v - 1}"
    elif inv_op == ">=":
        matcher = f"{v}.."
    elif inv_op == "<=":
        matcher = f"..{v}"
    ctx.file.append(f"execute if score {cpl.location} matches {matcher} run scoreboard players set {eid} 1")
    return CplScore(self.token, eid)


def num_and(self, _, cpl):
    if self.value == 0:
        return CplInt(self.token, 0)
    return cpl


def num_or(self, _, cpl):
    if self.value == 0:
        return cpl
    return CplInt(self.token, 1)


def nbt_set_add(self, ctx, cpl):
    if cpl.is_lit_eq(0):
        return self
    nbt_add(self, ctx, cpl).cache(ctx, nbt_loc=self.location, force="nbt")
    return self


def nbt_set_sub(self, ctx, cpl):
    if cpl.is_lit_eq(0):
        return self
    nbt_sub(self, ctx, cpl).cache(ctx, nbt_loc=self.location, force="nbt")
    return self


def nbt_set_mul(self, ctx, cpl):
    if cpl.is_lit_eq(1):
        return self
    nbt_mul(self, ctx, cpl).cache(ctx, nbt_loc=self.location, force="nbt")
    return self


def nbt_set_div(self, ctx, cpl):
    if cpl.is_lit_eq(0):
        raise_syntax_error("Division by zero", self.token)
    if cpl.is_lit_eq(1):
        return self
    nbt_div(self, ctx, cpl).cache(ctx, nbt_loc=self.location, force="nbt")
    return self


def nbt_set_mod(self, ctx, cpl):
    if cpl.is_lit_eq(0):
        raise_syntax_error("Modulo by zero", self.token)
    if cpl.is_lit_eq(1) and self.unique_type.type == "int":
        return self
    nbt_mod(self, ctx, cpl).cache(ctx, nbt_loc=self.location, force="nbt")
    return self


def nbt_set_pow(self, ctx, cpl):
    if cpl.is_lit_eq(0):
        return CplInt(self.token, 1)
    nbt_pow(self, ctx, cpl).cache(ctx, nbt_loc=self.location, force="nbt")
    return self


def nbt_add(self, ctx, cpl):
    if cpl.is_lit_eq(0):
        return self
    return self.cache(ctx, force="score")._set_add(ctx, cpl)


def nbt_sub(self, ctx, cpl):
    if cpl.is_lit_eq(0):
        return self
    return self.cache(ctx, force="score")._set_sub(ctx, cpl)


def nbt_mul(self, ctx, cpl):
    if cpl.is_lit_eq(1):
        return self
    if cpl.is_lit_eq(0):
        return cpl
    return self.cache(ctx, force="score")._set_mul(ctx, cpl)


def nbt_div(self, ctx, cpl):
    if cpl.is_lit_eq(0):
        raise_syntax_error("Division by zero", self.token)
    if cpl.is_lit_eq(1):
        return self
    return self.cache(ctx, force="score")._set_div(ctx, cpl)


def nbt_mod(self, ctx, cpl):
    if cpl.is_lit_eq(0):
        raise_syntax_error("Modulo by zero", self.token)
    if cpl.is_lit_eq(1) and self.unique_type.type == "int":
        return self
    return self.cache(ctx, force="score")._set_mod(ctx, cpl)


def nbt_pow(self, ctx, cpl):
    if cpl.is_lit_eq(0):
        return CplInt(self.token, 1)
    if cpl.is_lit_eq(1):
        return self
    return self.cache(ctx, force="score")._set_pow(ctx, cpl)


def nbt_eq_neq(self, ctx, cpl, is_eq):
    return self.cache(ctx, force="score")._eq_neq(ctx, cpl, is_eq)


def nbt_cmp(self, ctx, cpl, op):
    return self.cache(ctx, force="score")._cmp(ctx, cpl, op)


def nbt_and(self, ctx, cpl):
    return self.cache(ctx, force="score")._and(ctx, cpl)


def nbt_or(self, ctx, cpl):
    return self.cache(ctx, force="score")._or(ctx, cpl)


def make_num(cpl):
    cpl._add = num_add
    cpl._sub = num_sub
    cpl._mul = num_mul
    cpl._div = num_div
    cpl._mod = num_mod
    cpl._pow = num_pow
    cpl._eq_neq = num_eq_neq
    cpl._cmp = num_cmp
    cpl._and = num_and
    cpl._or = num_or


def make_nbt_num(cpl):
    cpl._set_add = nbt_set_add
    cpl._set_sub = nbt_set_sub
    cpl._set_mul = nbt_set_mul
    cpl._set_div = nbt_set_div
    cpl._set_mod = nbt_set_mod
    cpl._set_pow = nbt_set_pow
    cpl._add = nbt_add
    cpl._sub = nbt_sub
    cpl._mul = nbt_mul
    cpl._div = nbt_div
    cpl._mod = nbt_mod
    cpl._pow = nbt_pow
    cpl._eq_neq = nbt_eq_neq
    cpl._cmp = nbt_cmp
    cpl._and = nbt_and
    cpl._or = nbt_or



import json
from typing import List, Dict

from ._base import CompileTimeValue
from .float import CplFloat
from .int import CplInt
from .nbtobject import CplObjectNBT
from .string import CplString
from ..tokenizer import Token
from ..utils import CplDefObject


class CplObject(CompileTimeValue):
    def __init__(self, token=None, value=None, class_name=None):
        # type: (Token | None, Dict[str, CompileTimeValue] | None, str | None) -> None
        if value is None:
            value = dict()
        self.unique_type: CplDefObject
        super().__init__(token)
        self.value = value
        t = dict()
        for k, v in self.value.items():
            t[k] = v.unique_type
        self.unique_type = CplDefObject(t, class_name)

    def get_py_value(self):
        d = dict()
        for k, v in self.value.items():
            vl = v.get_py_value()
            if vl is None:
                return None
            d[k] = vl
        return d

    def _cache(self, ctx, score_loc=None, nbt_loc=None, force=None, force_t=None):
        if force == "score":
            return None
        py_val = self.get_py_value()
        if py_val is not None:
            ctx.file.append(f"data modify {nbt_loc} set value {json.dumps(py_val)}")
        else:
            init = []
            keys = []

            for k in self.value:
                v = self.value[k]
                v_py_val = v.get_py_value()
                if v_py_val is not None:
                    init.append(json.dumps(k) + ":" + json.dumps(v_py_val))
                else:
                    init.append(
                        json.dumps(k) + ":" + v.unique_type.get_sample_value()
                    )
                    keys.append(k)
            ctx.file.append(
                f"data modify {nbt_loc} set value {'{'}{','.join(init)}{'}'}"
            )

            for k in self.value:
                v = self.value[k]
                if k in keys:
                    v.cache(ctx, nbt_loc=f"{nbt_loc}.{k}", force="nbt")
        return CplObjectNBT(self.token, nbt_loc, self.unique_type)

    def _get_index(self, ctx, index: CompileTimeValue):
        if isinstance(index, CplInt) or isinstance(index, CplFloat) or isinstance(index, CplString):
            return self.value[str(index.value)]
        return self.cache(ctx).get_index(ctx, index)

    def _add(self, ctx, cpl):
        if isinstance(cpl, CplObject):
            if cpl.unique_type.class_name != self.unique_type.class_name:
                return None
            d = dict(self.value)
            d.update(cpl.value)
            return CplObject(self.token, d, self.unique_type.class_name)
        if not isinstance(cpl, CplObjectNBT):
            return None
        return self.cache(ctx)._add(ctx, cpl)

    def tellraw_object(self, ctx):
        py_v = self.get_py_value()
        if py_v is None:
            return self.cache(ctx).tellraw_object(ctx)
        return json.dumps(py_v)



from typing import List

from ..cpl._base import CompileTimeValue
from ..cpl.int import CplInt
from ..error import raise_syntax_error
from ..tokenizer import Token
from ..transpiler import FunctionDeclaration, TranspilerContext, add_lib


def lib_swap(ctx: TranspilerContext, args: List[CompileTimeValue], token: Token | None):
    if len(args) != 2:
        raise_syntax_error("Expected 2 arguments for <array>.swap()", token)
    arg0_cached = args[0].cache(ctx)
    args[0].compute(ctx, "=", args[1], token)
    args[1].compute(ctx, "=", arg0_cached, token)
    return CplInt(token, 0)


add_lib(FunctionDeclaration(
    type="python-cpl",
    name="swap",
    arguments=[],
    function=lib_swap
))



import json
from typing import List

from ..cpl.int import CplInt
from ..cpl.nbt import val_nbt, CplNBT
from ..cpl.object import CplObject
from ..error import raise_syntax_error
from ..tokenizer import GroupToken, cpl_def_from_tokens, Token
from ..transpiler import TranspilerContext, add_lib, CustomCplObject, FunctionDeclaration
from ..utils import VariableDeclaration, TokenType


def data_helper(s):
    if s[0] == "@":
        return "entity"
    if s[0] in list("0123456789~^-."):
        return "block"
    return "storage"


def lib_data_get(ctx: TranspilerContext, args: List[GroupToken], token: GroupToken):
    if len(args) != 2:
        raise_syntax_error(f"Expected 2 arguments for Data.get()", token)
    cpl_def = cpl_def_from_tokens(ctx.transpiler.classes, args[0].children)
    return val_nbt(token, data_helper(args[1].value) + " " + args[1].value, cpl_def)


def lib_data_set(ctx: TranspilerContext, args: List[GroupToken], token: GroupToken):
    if len(args) != 2:
        raise_syntax_error(f"Expected 2 arguments for Data.set()", token)
    return ctx.transpiler.tokens_to_cpl(ctx, args[1].children).cache(ctx, nbt_loc=args[0].value, force="nbt")


def lib_data_append(ctx: TranspilerContext, args: List[GroupToken], token: GroupToken):
    if len(args) != 2:
        raise_syntax_error(f"Expected 2 arguments for Data.append()", token)
    expr = ctx.transpiler.tokens_to_cpl(ctx, args[1].children)
    cmd = f"data modify {data_helper(args[0].value)} {args[0].value} append {expr.get_data_str(ctx)}"
    ctx.transpiler.run_cmd(ctx, Token(cmd, TokenType.POINTER, 0, len(cmd)))
    return expr


def lib_data_merge(ctx: TranspilerContext, args: List[GroupToken], token: GroupToken):
    if len(args) != 2:
        raise_syntax_error(f"Expected 2 arguments for Data.merge()", token)
    expr = ctx.transpiler.tokens_to_cpl(ctx, args[1].children)
    if expr.unique_type.type not in {"array", "object", "tuple"}:
        raise_syntax_error(f"Expected a merge-able object for the second argument of Data.merge()", token)
    if " " in args[0].value:
        cmd = f"data modify {data_helper(args[0].value)} {args[0].value} merge {expr.get_data_str(ctx)}"
    else:
        py_val = expr.get_py_value()
        if py_val is None or not isinstance(expr, CplObject):
            raise_syntax_error(
                f"Expected a literal object for the first argument of Data.merge() when the merging nbt location is an nbt root",
                token)
        cmd = f"data merge {data_helper(args[0].value)} {args[0].value} {json.dumps(py_val)}"
    ctx.transpiler.run_cmd(ctx, Token(cmd, TokenType.POINTER, 0, len(cmd)))
    return expr


def lib_data_remove(ctx: TranspilerContext, args: List[GroupToken], token: GroupToken):
    if len(args) != 1:
        raise_syntax_error(f"Expected 1 argument for Data.remove()", token)
    cmd = f"data remove {data_helper(args[0].value)} {args[0].value}"
    ctx.transpiler.run_cmd(ctx, Token(cmd, TokenType.POINTER, 0, len(cmd)))
    return CplInt(token, 0)


def lib_vset(ctx: TranspilerContext, args: List[GroupToken], token: GroupToken):
    if len(args) != 2:
        raise_syntax_error(f"Expected 2 arguments for vset()", token)
    val = ctx.transpiler.tokens_to_cpl(ctx, args[0].children)
    if not isinstance(val, CplNBT):
        raise_syntax_error("Expected an nbt pointer for the first argument of vset()", token)
    cmd = f"data modify {val.location} set value {args[1].value}"
    ctx.transpiler.run_cmd(ctx, Token(cmd, TokenType.POINTER, 0, len(cmd)))


def lib_mset(ctx: TranspilerContext, args: List[GroupToken], token: GroupToken):
    if len(args) != 2:
        raise_syntax_error(f"Expected 2 arguments for mset()", token)
    val = ctx.transpiler.tokens_to_cpl(ctx, args[0].children)
    if not isinstance(val, CplNBT):
        raise_syntax_error("Expected an nbt pointer for the first argument of mset()", token)
    cmd = f"data modify {val.location} set from {data_helper(args[1].value)} {args[1].value}"
    ctx.transpiler.run_cmd(ctx, Token(cmd, TokenType.POINTER, 0, len(cmd)))


add_lib(VariableDeclaration(
    name="Data",
    type=CustomCplObject({}, {
        "get": lib_data_get,
        "set": lib_data_set,
        "append": lib_data_append,
        "merge": lib_data_merge,
        "remove": lib_data_remove
    }),
    constant=True
), FunctionDeclaration(
    name="vset",
    type="python-raw",
    function=lib_vset
), FunctionDeclaration(
    name="mset",
    type="python-raw",
    function=lib_mset
))



import json
import os
import platform
import shutil
import sys
from argparse import ArgumentParser
from os import path
from time import sleep, time
from typing import Any

from .dp_ast import parse_str
from .transpiler import Transpiler
from .utils import VERSION_RADON, get_pack_format

BLACK = "\x1b[30m"
RED = "\x1b[31m"
GREEN = "\x1b[32m"
YELLOW = "\x1b[33m"
BLUE = "\x1b[34m"
MAGENTA = "\x1b[35m"
CYAN = "\x1b[36m"
WHITE = "\x1b[37m"
RESET = "\x1b[0m"
GRAY = "\x1b[90m"


# Usage: radon [build|watch] (-d="cwd")

class RadonArgumentParser(ArgumentParser):
    def __init__(self):
        super().__init__()
        self.add_argument("command", nargs="?", default="build", choices=["build", "watch"],
                          help="The command to run (build or watch)")
        self.add_argument("-d", default=os.getcwd(), type=str, help="sets the working directory")
        self.add_argument("-b", action="store_true", help="toggles debug mode")
        self.prog = "radon"

    def error(self, message):
        sys.stderr.write(RED)
        self.print_help(sys.stderr)
        sys.stderr.write(f"\nError: {message}\n")
        sys.stderr.write(RESET)
        sys.exit(2)


parser = RadonArgumentParser()
args = parser.parse_args()
cwd_list = args.d.split("|")
original_cwd = os.getcwd()


# real path
def pathr(pt):
    return path.realpath(pt)


def get_input(inp):
    try:
        return input(inp).strip()
    except KeyboardInterrupt:
        print(f"\n{RED}Cancelling...{RESET}")
        sys.exit(0)


def clear():
    if platform.system() == "Windows":
        os.system("cls")
    else:
        os.system("clear")


def read_config():
    with open(f"./radon.json", "r") as file:
        res = json.loads(file.read())
        if "useLock" not in res:
            res["useLock"] = False
        return res


def listdir_recursive(pt):
    for root, _, files in os.walk(pt):
        for file in files:
            yield pathr(root + "/" + file)


def empty_dir_recursive(pt):
    if path.exists(pt) and path.isdir(pt):
        for f in os.listdir(pt):
            empty_dir_recursive(pathr(pt + "/" + f))
        if len(os.listdir(pt)) == 0:
            os.rmdir(pt)
            empty_dir_recursive(pathr(pt + "/../"))


def build_dir():
    init_dir()

    config = read_config()

    if not path.exists(config["main"]):
        return "The main file does not exist: " + config["main"] + ", cwd: " + os.getcwd()

    with open(pathr(config["main"]), "r") as file:
        code = file.read()

    try:
        (statements, macros) = parse_str(code)
        transpiler = Transpiler(
            statements=statements,
            macros=macros,
            pack_namespace=config["namespace"],
            pack_description=config["description"],
            pack_format=get_pack_format(config["format"]),
            main_dir=config["main"] + "/../",
            main_file_path=config["main"],
            debug_mode=args.b)
    except SyntaxError as e:
        return str(e)
    except Exception as e:
        raise e

    dp_files = transpiler.get_datapack_files()

    out_folders = (
        config["outFolder"]
        if isinstance(config["outFolder"], list)
        else [config["outFolder"]]
    )

    lock = ""

    if config["useLock"] and path.exists(f"./radon.lock"):
        with open(f"./radon.lock", "r") as file:
            lock = file.read().strip("\n")

    rm_files = lock.split("\n") if lock else []

    for out_folder in out_folders:
        if config["useLock"]:
            for f in rm_files:
                f = f"{out_folder}/{f}"
                if not path.exists(f):
                    continue
                if path.isfile(f):
                    os.remove(f)
                dr = path.dirname(f)
                empty_dir_recursive(dr)
        else:
            shutil.rmtree(out_folder + "/data", ignore_errors=True)
            if path.exists(out_folder + "/pack.mcmeta"):
                os.remove(out_folder + "/pack.mcmeta")
            os.makedirs(out_folder + "/data", exist_ok=True)

    for out_folder in out_folders:
        if "data" in config and path.exists(config["data"]) and path.isdir(config["data"]):
            shutil.copytree(config["data"], out_folder + "/data", dirs_exist_ok=True)
        for pt in dp_files:
            file_name = pt
            pt = f"{out_folder}/{pt}"
            os.makedirs(pathr(pt + "/../"), exist_ok=True)
            with open(pt, "w") as file:
                file.write(dp_files[file_name])

    if config["useLock"]:
        with open(f"./radon.lock", "w") as file:
            file.write("\n".join(dp_files.keys()))


def build():
    res = []
    for cwd in cwd_list:
        os.chdir(original_cwd)
        if not path.exists(cwd):
            print(f"{RED}Directory {cwd} does not exist!{RESET}")
            exit(1)
        os.chdir(cwd)
        s = build_dir()
        if isinstance(s, str):
            res.append(s)
    return "\n".join(res) if len(res) > 0 else None


def init_dir():
    if path.exists(f"./radon.json"):
        return
    print(f"{BLUE}No config file found, generating radon.json")

    pack_namespace = None
    while not pack_namespace:
        pack_namespace = get_input(f"{CYAN}*Namespace: {RESET}")

    if not pack_namespace:
        print(f"{RED}Cancelling...{RESET}")
        sys.exit(0)

    pack_desc = get_input(f"{CYAN}Description: {RESET}")

    pack_format = None

    while not get_pack_format(pack_format):
        pack_format = get_input(f"{CYAN}*Pack format or Minecraft version: {RESET}")

    pack_main = get_input(f"{CYAN}Main file(src/main.rn): {RESET}") or "src/main.rn"
    pack_data = get_input(f"{CYAN}Data folder(src/data): {RESET}") or "src/data"
    pack_output_folder = get_input(f"{CYAN}Output folder(.): {RESET}") or "."

    with open("./radon.json", "w") as file:
        file.write(
            json.dumps(
                {
                    "namespace": pack_namespace,
                    "description": pack_desc,
                    "format": pack_format,
                    "main": pack_main,
                    "data": pack_data,
                    "outFolder": pack_output_folder,
                    "useLock": False,
                },
                indent=2,
            )
        )

    if not path.exists(pack_main):
        os.makedirs(path.dirname(pack_main), exist_ok=True)
        with open(pack_main, "w") as file:
            file.write("print('Hello, world!')")
    if not path.exists(pack_data):
        os.makedirs(pack_data, exist_ok=True)

    print(f"{GREEN}Datapack's workplace is ready!{RESET}")
    build_for_watch()
    sys.exit(0)


def init():
    for cwd in cwd_list:
        os.chdir(original_cwd)
        if not path.exists(cwd):
            print(f"{RED}Directory {cwd} does not exist!{RESET}")
            exit(1)
        os.chdir(cwd)
        init_dir()


def watch_snapshot():
    files = []
    files += map(
        lambda x: path.join("./", x),
        list(listdir_recursive("./src")) + ["radon.json"]
    )
    return {f: path.getmtime(f) for f in files}


def build_for_watch():
    start = time()
    built = build()
    if isinstance(built, str):
        print(RED + built + RESET)
    else:
        took = time() - start
        print(f"{GREEN}Datapack has been built!{GRAY} ({took:.5f}s){RESET}")


files_snapshot: Any = None


def check_for_changes():
    global files_snapshot
    new_snapshot = watch_snapshot()
    old_snapshot = files_snapshot

    if new_snapshot == old_snapshot:
        return

    for file in new_snapshot:
        if file not in old_snapshot:
            print(f"{GRAY}File created: {path.realpath(file)}{RESET}")
        elif new_snapshot[file] != old_snapshot[file]:
            print(f"{GRAY}File modified: {path.realpath(file)}{RESET}")

    for file in old_snapshot:
        if file not in new_snapshot:
            print(f"{GRAY}File deleted: {path.realpath(file)}{RESET}")

    build_for_watch()

    files_snapshot = new_snapshot


def main():
    global files_snapshot
    print(f"{YELLOW}Radon v{VERSION_RADON}{RESET}")
    print("")
    print(f"{CYAN}Current Directory | {" | ".join(cwd_list)}{RESET}")
    print("")

    init()

    if args.command == "build":
        start = time()
        built = build()
        if isinstance(built, str):
            print(RED + built + RESET)
            sys.exit(1)

        took = time() - start

        print(f"{GREEN}Datapack has been built!{GRAY} ({took:.5f}s){RESET}")
        sys.exit(0)

    files_snapshot = watch_snapshot()

    print(f"{GREEN}Watching files...{RESET}")
    print(f"{CYAN}Press CTRL + C to stop watching{RESET}")
    build_for_watch()
    try:
        while True:
            sleep(0.3)
            check_for_changes()
    except KeyboardInterrupt:
        print(f"{RED}Stopped watching{RESET}")

    sys.exit(0)


if __name__ == "__main__":
    main()



from typing import Union, List

from ._base import CompileTimeValue
from .nbt import CplNBT
from ..tokenizer import Token
from ..utils import FLOAT_TYPE, FLOAT_PREC


class CplFloatNBT(CplNBT):
    def __init__(self, token: Union[Token, None], location: str):
        super().__init__(token, location, FLOAT_TYPE)

    def _cache(self, ctx, score_loc=None, nbt_loc=None, force=None, force_t=None):
        if force == "score":
            t = "float" or force_t
            if t == "int":
                ctx.file.append(f"execute store result score {score_loc} run data get {self.location}")
            else:
                ctx.file.append(f"execute store result score {score_loc} run data get {self.location} {FLOAT_PREC}")
            return CplScore(self.token, score_loc, t)
        return super()._cache(ctx, score_loc, nbt_loc, force, force_t)

    def _get_index(self, ctx, index: CompileTimeValue):
        return None

    def _get_slice(self, ctx, index1, index2, index3, token):
        return None

    def _call_index(self, ctx, index: str, arguments: List[CompileTimeValue], token):
        return self.cache(ctx, force="score")._call_index(ctx, index, arguments, token)


from ._number import make_nbt_num
from .score import CplScore

make_nbt_num(CplFloatNBT)



import json
from typing import List

from ..cpl._base import CompileTimeValue
from ..cpl.int import CplInt
from ..cpl.object import CplObject
from ..error import raise_syntax_error
from ..transpiler import add_lib, TranspilerContext
from ..utils import VariableDeclaration, get_uuid


class RecipeObject(CplObject):
    def _call_index(self, ctx: TranspilerContext, index: str, arguments: List[CompileTimeValue], token):
        tr = ctx.transpiler
        if index == "add":
            if len(arguments) < 1:
                raise_syntax_error("Recipe.table() requires at least 1 argument", token)
            recipe = arguments[0]
            if not isinstance(recipe, CplObject):
                raise_syntax_error("Expected an object for the first argument of Recipe.table()", token)
            recipe = recipe.get_py_value()
            if not recipe:
                raise_syntax_error("Expected the table to be a literal object", token)
            tr.dp_files[f"recipe{tr.s}/recipe_{get_uuid()}.json"] = json.dumps(recipe, indent=2)
            return CplInt(token, 0)
        return None


add_lib(VariableDeclaration(
    name="Recipe",
    type=RecipeObject(),
    constant=True
))



from typing import List

from ..cpl._base import CompileTimeValue
from ..cpl.score import CplScore
from ..error import raise_syntax_error
from ..transpiler import FunctionDeclaration, TranspilerContext, add_lib
from ..utils import FLOAT_PREC


def lib_time(ctx: TranspilerContext, args: List[CompileTimeValue], token):
    if len(args) != 0:
        raise_syntax_error("time() takes no arguments", token)
    tr = ctx.transpiler
    tr.files["__load__"].append("scoreboard players set __time__time__ __temp__ 0")
    tr.tick_file.append("scoreboard players add __time__time__ __temp__ 1")
    return CplScore(token, "__time__time__ __temp__", "int")


def lib_ftime(ctx: TranspilerContext, args: List[CompileTimeValue], token):
    if len(args) != 0:
        raise_syntax_error("ftime() takes no arguments", token)
    tr = ctx.transpiler
    tr.files["__load__"].append("scoreboard players set __ftime__time__ __temp__ 0")
    tr.tick_file.append(f"scoreboard players add __ftime__time__ __temp__ {int(FLOAT_PREC / 20)}")
    return CplScore(token, "__ftime__time__ __temp__", "float")


add_lib(FunctionDeclaration(
    type="python-cpl",
    name="time",
    function=lib_time,
), FunctionDeclaration(
    type="python-cpl",
    name="ftime",
    function=lib_ftime,
))



from typing import Union, List

from ._base import CompileTimeValue
from .nbt import CplNBT
from ..tokenizer import Token
from ..utils import INT_TYPE


class CplIntNBT(CplNBT):
    def __init__(self, token: Union[Token, None], location: str):
        super().__init__(token, location, INT_TYPE)

    def _cache(self, ctx, score_loc=None, nbt_loc=None, force=None, force_t=None):
        if force == "score":
            ctx.file.append(f"execute store result score {score_loc} run data get {self.location}")
            t = "int"
            if force_t == "float":
                ctx.file.append(f"scoreboard players operation {score_loc} *= FLOAT_PREC __temp__")
                t = "float"
            return CplScore(self.token, score_loc, t)
        return super()._cache(ctx, score_loc, nbt_loc, force, force_t)

    def _call_index(self, ctx, index: str, arguments: List[CompileTimeValue], token):
        return self.cache(ctx, force="score")._call_index(ctx, index, arguments, token)


from .score import CplScore
from ._number import make_nbt_num

make_nbt_num(CplIntNBT)



from setuptools import setup, find_packages
from radon.utils import VERSION_RADON

setup(
    name="radonmc",
    version=VERSION_RADON,
    packages=find_packages(),
    package_data={
        "radon": ["builtin/*", "cpl/*"],
    },
    entry_points={
        "console_scripts": [
            "radon=radon.__main__:main",
        ],
    },
    install_requires=[],
    author="OguzhanUmutlu",
    description="Radon is a programming language that gets compiled to Minecraft: Java Edition's mcfunction files.",
    long_description=open("../README.md").read(),
    long_description_content_type="text/markdown",
    url="https://github.com/OguzhanUmutlu/radon",
    license="MIT"
)



from typing import Union

from ._base import CompileTimeValue
from ..tokenizer import Token
from ..utils import get_uuid, CplDef, CplDefObject, CplDefArray


class CplNBT(CompileTimeValue):
    def __init__(self, token: Union[Token, None], location: str, type: CplDef):
        super().__init__(token)
        self.location = location
        self.unique_type = type

    def _cache(self, ctx, score_loc=None, nbt_loc=None, force=None, force_t=None):
        if force == "score":
            return None
        if not nbt_loc:
            nbt_loc = f"storage temp _{get_uuid()}"
        ctx.file.append(f"data modify {nbt_loc} set from {self.location}")
        return val_nbt(self.token, nbt_loc, self.unique_type)

    def _set(self, ctx, cpl):
        if self.unique_type != cpl.unique_type:
            return None
        cpl.cache(ctx, nbt_loc=self.location, force="nbt")
        return self

    def tellraw_object(self, ctx):
        ls = self.location.split(" ")
        ls1 = ls[1]
        ls2 = " ".join(ls[2:])
        if ls[0] == "block":
            ls1 = ls[1] + " " + ls[2] + " " + ls[3]
            ls2 = " ".join(ls[4:])
        return '{"' + ls[0] + '":"' + ls1 + '","nbt":"' + ls2 + '"}'

    def get_data_str(self, ctx):
        return f"from {self.location}"


def val_nbt(
        token: Union[Token, None],
        location: str,
        u_type: CplDef
) -> CplNBT:
    ut = u_type.type
    if ut == "int":
        return CplIntNBT(token, location)
    elif ut == "float":
        return CplFloatNBT(token, location)
    elif ut == "string":
        return CplStringNBT(token, location)
    elif isinstance(u_type, CplDefArray):
        return CplArrayNBT(token, location, u_type)
    elif isinstance(u_type, CplDefObject):
        return CplObjectNBT(token, location, u_type)
    raise ValueError("Invalid type")


def object_get_index_nbt(ctx, content_type, cpl, index):
    eid = f"storage object_index _{get_uuid()}"
    file_name = ctx.transpiler.get_temp_file_name("$data modify {eid} set from $(_0).$(_1)")

    cpl.cache(ctx, force="nbt", nbt_loc="object_get_index _0")
    index.cache(ctx, force="nbt", nbt_loc="object_get_index _1")
    ctx.file.append(f"function {ctx.transpiler.pack_namespace}:{file_name} with storage object_get_index")
    return val_nbt(cpl.token, eid, content_type)


from .nbtarray import CplArrayNBT
from .nbtfloat import CplFloatNBT
from .nbtint import CplIntNBT
from .nbtobject import CplObjectNBT
from .nbtstring import CplStringNBT



from ..cpl.int import CplInt
from ..transpiler import add_lib
from ..utils import VariableDeclaration

add_lib(
    VariableDeclaration(
        name="true",
        type=CplInt(None, 1),
        constant=True
    ),
    VariableDeclaration(
        name="false",
        type=CplInt(None, 0),
        constant=True
    ),
    VariableDeclaration(
        name="null",
        type=CplInt(None, 0),
        constant=True
    ))



from typing import List

from ..cpl.selector import CplSelector
from ..cpl._base import CompileTimeValue
from ..cpl.int import CplInt
from ..error import raise_syntax_error
from ..transpiler import FunctionDeclaration, TranspilerContext, add_lib


def _lib_print(ctx: TranspilerContext, args: List[CompileTimeValue], token, prefix: str, name: str):
    if len(args) < 1:
        raise_syntax_error(f"Invalid arguments. Expected usage: {name}(selector, ...anything)", token)
    if not isinstance(args[0], CplSelector):
        args.insert(0, CplSelector(token, "@a"))
    prefix = prefix.replace("@", args[0].value)
    if len(args) == 2:
        ctx.file.append(f"{prefix} {args[1].tellraw_object(ctx)}")
    else:
        ctx.file.append(f"{prefix} [{','.join(x.tellraw_object(ctx) for x in args[1:])}]")
    return CplInt(token, 0)


def lib_print(ctx: TranspilerContext, args: List[CompileTimeValue], token):
    return _lib_print(ctx, args, token, "tellraw @", "print")


def lib_print_title(ctx: TranspilerContext, args: List[CompileTimeValue], token):
    return _lib_print(ctx, args, token, "title @ title", "printTitle")


def lib_print_subtitle(ctx: TranspilerContext, args: List[CompileTimeValue], token):
    return _lib_print(ctx, args, token, "title @ subtitle", "printSubtitle")


def lib_print_actionbar(ctx: TranspilerContext, args: List[CompileTimeValue], token):
    return _lib_print(ctx, args, token, "title @ actionbar", "printActionbar")


add_lib(FunctionDeclaration(
    type="python-cpl",
    name="print",
    arguments=[],
    function=lib_print
), FunctionDeclaration(
    type="python-cpl",
    name="printTitle",
    arguments=[],
    function=lib_print_title
), FunctionDeclaration(
    type="python-cpl",
    name="printSubtitle",
    arguments=[],
    function=lib_print_subtitle
), FunctionDeclaration(
    type="python-cpl",
    name="printActionbar",
    arguments=[],
    function=lib_print_actionbar
))
